﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Runtime.ConstrainedExecution;
using System.IO;

namespace Prelude.Internal
{



    /// <summary>
    /// Restart Manager の部分的な .NET ラッパー
    /// </summary>
    /// <remarks>http://msdn.microsoft.com/msdnmag/issues/07/04/NETMatters/default.aspx?loc=jp</remarks>
    static class RestartMgr
    {
        private const int RmGetListLoopCount = 3;
        private const int RmGetListInitialArraySize = 16;

        public static Process[] GetProcessesByWhichFilesInUse(params FileInfo[] files)
        {
            using(SafeRmSessionHandle session = new SafeRmSessionHandle())
            {
                session.StartSession();

                RegisterFiles(session, files);

                List<Process> procList = new List<Process>();
                foreach (RM_PROCESS_INFO procInfo in GetProcessInfoList(session))
                {
                    RM_UNIQUE_PROCESS uniqueProc = procInfo.Process;
                    Process proc = Process.GetProcessById(uniqueProc.dwProcessId);

                    bool same = true;

                    try
                    {
                        ulong highPart = unchecked((uint)uniqueProc.ProcessStartTime.dwHighDateTime);
                        ulong lowPart = unchecked((uint)uniqueProc.ProcessStartTime.dwLowDateTime);
                        long filetime = checked((long)((highPart << 32) + lowPart));
                        DateTime startTime = DateTime.FromFileTime(filetime);
                        same &= (proc.StartTime == startTime);
                    }
                    catch(Win32Exception)
                    {
                        // Ignore error
                    }

                    if (same)
                    {
                        procList.Add(proc);
                    }
                    else
                    {
                        proc.Dispose();
                    }
                }

                return procList.ToArray();
            }
        }

        static RM_PROCESS_INFO[] GetProcessInfoList(SafeRmSessionHandle session)
        {
            uint numProcInfoNeeded = RmGetListInitialArraySize;
            int loopLimit = RmGetListLoopCount;

            while (true)
            {
                if (loopLimit <= 0)
                    throw new Exception("RmGetList reached to maximum loop count");

                RM_PROCESS_INFO[] processInfo = new RM_PROCESS_INFO[numProcInfoNeeded];
                uint numProcInfo = (uint)processInfo.Length;
                RM_REBOOT_REASON reasons = RM_REBOOT_REASON.RmRebootReasonNone;

                RmGetListResult result = SafeRmSessionHandle.RmGetList(session.SessionHandle, out numProcInfoNeeded, ref numProcInfo, processInfo, ref reasons);

                switch (result)
                {
                    case RmGetListResult.ERROR_SUCCESS:
                        {
                            RM_PROCESS_INFO[] retbuf = new RM_PROCESS_INFO[numProcInfo];
                            Array.Copy(processInfo, retbuf, numProcInfo);
                            return retbuf;
                        }
                    case RmGetListResult.ERROR_MORE_DATA:
                        --loopLimit;
                        if (loopLimit == 1) numProcInfoNeeded *= 2;
                        break;
                    default:
                        throw new Win32Exception((int)result, "RmEndSession Failed");
                }
            }
        }

        static void RegisterFiles(SafeRmSessionHandle session, IEnumerable<FileInfo> files)
        {
            List<string> filePaths = new List<string>();
            foreach (FileInfo file in files)
            {
                filePaths.Add(file.FullName);
            }

            string[] pathStrings = new string[filePaths.Count];
            filePaths.CopyTo(pathStrings, 0);

            RmRegisterResourcesResult result = SafeRmSessionHandle.RmRegisterResources(session.SessionHandle, (uint)pathStrings.Length, pathStrings, 0, null, 0, null);
            if (result != RmRegisterResourcesResult.ERROR_SUCCESS)
            {
                throw new Win32Exception((int)result, "RmRegisterResources Failed");
            }
        }

        internal const int RM_SESSION_KEY_LEN = 128; // = Marshal.SizeOf(typeof(Guid))
        internal const int CCH_RM_SESSION_KEY = RM_SESSION_KEY_LEN * 2;
        internal const int CCH_RM_MAX_APP_NAME = 255;
        internal const int CCH_RM_MAX_SVC_NAME = 63;
        internal const int RM_INVALID_TS_SESSION = -1;
        internal const int RM_INVALID_PROCESS = -1;

        [StructLayout(LayoutKind.Sequential)]
        private struct RM_UNIQUE_PROCESS
        {
            public int dwProcessId;
            public System.Runtime.InteropServices.ComTypes.FILETIME ProcessStartTime;
        }

        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
        private struct RM_PROCESS_INFO
        {
            public RM_UNIQUE_PROCESS Process;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = CCH_RM_MAX_APP_NAME + 1)]
            public string strAppName;
            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = CCH_RM_MAX_SVC_NAME + 1)]
            public string strServiceShortName;
            public RM_APP_TYPE ApplicationType;
            public RM_APP_STATUS AppStatus;
            public uint TSSessionId;
            [MarshalAs(UnmanagedType.Bool)]
            public bool bRestartable;
        }

        sealed class SafeRmSessionHandle : CriticalFinalizerObject, IDisposable
        {
            private int _sessionHandle = RestartMgr.RM_INVALID_TS_SESSION;

            public int SessionHandle
            {
                get { return _sessionHandle; }
                set { _sessionHandle = value; }
            }

            public bool IsSessionOpen
            {
                get { return SessionHandle != RestartMgr.RM_INVALID_TS_SESSION; }
            }

            [ReliabilityContract(Consistency.MayCorruptProcess, Cer.MayFail)]
            ~SafeRmSessionHandle()
            {
                ReleaseInternal();
            }

            public void StartSession()
            {
                if (IsSessionOpen)
                    throw new Exception("The RestartMgr session has already been opened.");

                RmStartSessionResult result = RmStartSession(out _sessionHandle, 0, Guid.NewGuid().ToString("N"));
                if(result != RmStartSessionResult.ERROR_SUCCESS)
                    throw new Win32Exception((int) result, "RmStartSession Failed");
            }

            [ReliabilityContract(Consistency.MayCorruptProcess, Cer.MayFail)]
            private void ReleaseInternal()
            {
                if (IsSessionOpen)
                {
                    RmEndSessionResult result = RmEndSession(_sessionHandle);
                    if (result != RmEndSessionResult.ERROR_SUCCESS)
                    {
                        throw new Win32Exception((int)result, "RmEndSession Failed");
                    }
                    _sessionHandle = RestartMgr.RM_INVALID_TS_SESSION;
                }
            }

            [ReliabilityContract(Consistency.MayCorruptProcess, Cer.MayFail)]
            public void Dispose()
            {
                ReleaseInternal();
                GC.SuppressFinalize(this);
            }

            [DllImport("rstrtmgr.dll", CharSet = CharSet.Unicode)]
            internal static extern RmStartSessionResult RmStartSession(
                out int pSessionHandle,
                int dwSessionFlags,
                [MarshalAs(UnmanagedType.LPWStr, SizeConst = CCH_RM_SESSION_KEY + 1)]
                string strSessionKey);

            [DllImport("rstrtmgr.dll")]
            internal static extern RmEndSessionResult RmEndSession(int pSessionHandle);

            [DllImport("rstrtmgr.dll", CharSet = CharSet.Unicode)]
            internal static extern RmRegisterResourcesResult RmRegisterResources(
                int pSessionHandle,
                UInt32 nFiles,
                string[] rgsFilenames,
                UInt32 nApplications, RM_UNIQUE_PROCESS[] rgApplications, UInt32 nServices, string[] rgsServiceNames);

            [DllImport("rstrtmgr.dll")]
            internal static extern RmGetListResult RmGetList(
                int dwSessionHandle,
                out uint pnProcInfoNeeded,
                ref uint pnProcInfo,
                [In, Out] RM_PROCESS_INFO[] rgAffectedApps,
                ref RM_REBOOT_REASON lpdwRebootReasons);
        }

        private enum RM_APP_TYPE
        {
            /// <summary>
            /// Application type cannot be classified in known categories
            /// </summary>
            RmUnknownApp = 0,
            /// <summary>
            /// Application is a windows application that displays a top-level window
            /// </summary>
            RmMainWindow = 1,
            /// <summary>
            /// Application is a windows app but does not display a top-level window
            /// </summary>
            RmOtherWindow = 2,
            /// <summary>
            /// Application is an NT service
            /// </summary>
            RmService = 3,
            /// <summary>
            /// Application is Explorer
            /// </summary>
            RmExplorer = 4,
            /// <summary>
            /// Application is Console application
            /// </summary>
            RmConsole = 5,
            /// <summary>
            /// Application is critical system process where a reboot is required to restart
            /// </summary>
            RmCritical = 1000
        }

        [Flags]
        enum RM_APP_STATUS : uint
        {
            /// <summary>
            /// The application is in a state that is not described by any other enumerated state. 
            /// </summary>
            RmStatusUnknownRmStatusUnknown = 0x0,
            /// <summary>
            /// The application is currently running. 
            /// </summary>
            RmStatusRunningRmStatusRunning = 0x1,
            /// <summary>
            /// The Restart Manager has stopped the application. 
            /// </summary>
            RmStatusStoppedRmStatusStoppedOther = 0x2,
            /// <summary>
            /// An action outside the Restart Manager has stopped the application. 
            /// </summary>
            RmStatusStoppedOtherRmStatusStopped = 0x4,
            /// <summary>
            /// The Restart Manager has restarted the application. 
            /// </summary>
            RmStatusRestartedRmStatusRestarted = 0x8,
            /// <summary>
            /// The Restart Manager encountered an error when stopping the application. 
            /// </summary>
            RmStatusErrorOnStopRmStatusErrorOnStop = 0x10,
            /// <summary>
            /// The Restart Manager encountered an error when restarting the application. 
            /// </summary>
            RmStatusErrorOnRestartRmStatusErrorOnRestart = 0x20,
            /// <summary>
            /// THe Restart Manager shutdown actions have been modified. 
            /// </summary>
            RmStatusShutdownMaskedRmStatusShutdownMasked = 0x40,
            /// <summary>
            /// THe Restart Manager restart actions have been modified
            /// </summary>
            RmStatusRestartMaskedRmStatusRestartMasked = 0x80
        }


        [Flags]
        enum RM_REBOOT_REASON : uint
        {
            /// <summary>
            /// Reboot not required
            /// </summary>
            RmRebootReasonNone = 0x0,
            /// <summary>
            /// Current user does not have permission to shut down one or more detected processes
            /// </summary>
            RmRebootReasonPermissionDenied = 0x1,
            /// <summary>
            /// One or more processes are running in another TS session.
            /// </summary>
            RmRebootReasonSessionMismatch = 0x2,
            /// <summary>
            /// A critical process has been detected
            /// </summary>
            RmRebootReasonCriticalProcess = 0x4,
            /// <summary>
            /// A critical service has been detected
            /// </summary>
            RmRebootReasonCriticalService = 0x8,
            /// <summary>
            /// The current process has been detected
            /// </summary>
            RmRebootReasonDetectedSelf = 0x10,
        }


        enum RmStartSessionResult : uint
        {
            /// <summary>
            /// The function completed successfully. 
            /// </summary>
            ERROR_SUCCESS = 0,
            /// <summary>
            /// A Restart Manager function could not obtain a Registry write mutex in the allotted time.
            /// A system restart is recommended because further use of the Restart Manager is likely to fail. 
            /// </summary>
            ERROR_SEM_TIMEOUT = 121,
            /// <summary>
            /// One or more arguments are not correct.
            /// This error value is returned by the Restart Manager function if a NULL pointer or 0 is passed in a parameter that requires a non-null and non-zero value. 
            /// </summary>
            ERROR_BAD_ARGUMENTS = 160,
            /// <summary>
            /// The maximum number of sessions has been reached.
            /// </summary>
            ERROR_MAX_SESSIONS_REACHED = 353,
            /// <summary>
            /// The system cannot write to the specified device.
            /// </summary>
            ERROR_WRITE_FAULT = 29,
            /// <summary>
            /// A Restart Manager operation could not complete because not enough memory was available. 
            /// </summary>
            ERROR_OUTOFMEMORY = 14,
        }

        enum RmEndSessionResult : uint
        {
            /// <summary>
            /// The function completed successfully. 
            /// </summary>
            ERROR_SUCCESS = 0,
            /// <summary>
            /// A Restart Manager function could not obtain a Registry write mutex in the allotted time.
            /// A system restart is recommended because further use of the Restart Manager is likely to fail. 
            /// </summary>
            ERROR_SEM_TIMEOUT = 121,
            /// <summary>
            /// One or more arguments are not correct.
            /// This error value is returned by the Restart Manager function if a NULL pointer or 0 is passed in a parameter that requires a non-null and non-zero value. 
            /// </summary>
            ERROR_BAD_ARGUMENTS = 160,
            /// <summary>
            /// An operation was unable to read or write to the registry.
            /// </summary>
            ERROR_WRITE_FAULT = 29,
            /// <summary>
            /// A Restart Manager operation could not complete because not enough memory was available. 
            /// </summary>
            ERROR_OUTOFMEMORY = 14,
            /// <summary>
            /// No Restart Manager session exists for the handle supplied.
            /// </summary>
            ERROR_INVALID_HANDLE = 6,
        }

        enum RmRegisterResourcesResult : uint
        {
            /// <summary>
            /// The resources specified have been registered. 
            /// </summary>
            ERROR_SUCCESS = 0,
            /// <summary>
            /// A Restart Manager function could not obtain a Registry write mutex in the allotted time.
            /// A system restart is recommended because further use of the Restart Manager is likely to fail. 
            /// </summary>
            ERROR_SEM_TIMEOUT = 121,
            /// <summary>
            /// One or more arguments are not correct.
            /// This error value is returned by the Restart Manager function if a NULL pointer or 0 is passed in a parameter that requires a non-null and non-zero value. 
            /// </summary>
            ERROR_BAD_ARGUMENTS = 160,
            /// <summary>
            /// An operation was unable to read or write to the registry.
            /// </summary>
            ERROR_WRITE_FAULT = 29,
            /// <summary>
            /// A Restart Manager operation could not complete because not enough memory was available. 
            /// </summary>
            ERROR_OUTOFMEMORY = 14,
            /// <summary>
            /// No Restart Manager session exists for the handle supplied.
            /// </summary>
            ERROR_INVALID_HANDLE = 6,
        }

        enum RmGetListResult
        {
            /// <summary>
            /// The function completed successfully. 
            /// </summary>
            ERROR_SUCCESS = 0,
            /// <summary>
            /// This error value is returned by the RmGetList function if the rgAffectedApps buffer is too small to hold all application information in the list. 
            /// </summary>
            ERROR_MORE_DATA = 234,
            /// <summary>
            /// The current operation is cancelled by user. 
            /// </summary>
            ERROR_CANCELLED = 1223,
            /// <summary>
            /// A Restart Manager function could not obtain a Registry write mutex in the allotted time.
            /// A system restart is recommended because further use of the Restart Manager is likely to fail. 
            /// </summary>
            ERROR_SEM_TIMEOUT = 121,
            /// <summary>
            /// One or more arguments are not correct.
            /// This error value is returned by the Restart Manager function if a NULL pointer or 0 is passed in a parameter that requires a non-null and non-zero value. 
            /// </summary>
            ERROR_BAD_ARGUMENTS = 160,
            /// <summary>
            /// An operation was unable to read or write to the registry. 
            /// </summary>
            ERROR_WRITE_FAULT = 29,
            /// <summary>
            /// A Restart Manager operation could not complete because not enough memory was available. 
            /// </summary>
            ERROR_OUTOFMEMORY = 14,
            /// <summary>
            /// No Restart Manager session exists for the handle supplied. 
            /// </summary>
            ERROR_INVALID_HANDLE = 6,
        }
    }
}

